local _M = {_NAME = (...), _PACKAGE = string.gsub((...), "^(.-)[^%.]*$", "%1")}

local DEFAULT_RANGE = 10


---------------------------------------------------------------------
-- Local functions
---------------------------------------------------------------------

--- Normalize addEsp() and removeEsp() input.
-- @param esp_types A table of ESP types and/or ESP type = range pairs OR a bare string containing
--                  a single ESP type.  All types without explicit ranges will use the default range.
-- @return A normalized list of ESP type = range pairs.
local function normalizeEspTypesAndRanges(esp_types)
	local esp_types_list = {}

	if type(esp_types) == "string" then
		esp_types_list[_M.normalizeEspType(esp_types)] = DEFAULT_RANGE
	elseif type(esp_types) == "table" then
		for esp_type, range in pairs(esp_types) do
			if type(esp_type) == "number" then
				esp_types_list[_M.normalizeEspType(range)] = DEFAULT_RANGE
			else
				esp_types_list[_M.normalizeEspType(esp_type)] = range
			end
		end
	else
		error("ESP error - normalizeEspTypesAndRanges(): only strings and tables are supported")
	end
	
	return esp_types_list
end

--- Add or remove ESP types and ranges to/from the active ESP pool.
-- @param esp The Actor's active ESP table.
-- @param esp_types A table of ESP types and/or ESP type = range pairs OR a bare string containing
--                  a single ESP type.  All types without explicit ranges will use the default range.
-- @param rt_function A function that changes the range table(s) of the given type(s).
-- @return The number of ESP effects successfully added/removed.
local function changeTypesAndRanges(esp, esp_types, rt_function)
	local changes = 0
	
	for esp_type, range in pairs(esp_types) do
		if rt_function(esp, esp_type, range) then changes = changes + 1 end
	end

	return changes
end

--- Add a range to an ESP type's range list.
-- @param esp The Actor's active ESP table.
-- @param esp_type A string containing the normalized ESP type to update (and add to the
--                 active pool if not there).
-- @param range The range to add to the type's range list.
-- @return A boolean indicating the success of the range addition.
local function addRange(esp, esp_type, range)
	range = range or DEFAULT_RANGE

	-- Add the ESP type if it isn't there
	esp[esp_type] = esp[esp_type] or {}
	local range_list = esp[esp_type]

	-- Find the index where the range should be inserted...
	local i = 1	
	while i <= #range_list and range_list[i] > range do i = i + 1 end

	-- ... move the shorter ranges down the list...
	for j = #range_list, i, -1 do
		range_list[j + 1] = range_list[j]
	end
	
	-- ... and insert it!
	range_list[i] = range
	
	return true
end

--- Remove a range from an ESP type's range list.
-- @param esp The Actor's active ESP table.
-- @param esp_type A string containing the normalized ESP type to update (and remove
--                 from the active pool if empty).
-- @param range The range to remove from the type's range list.
-- @return A boolean indicating the success of the range removal.
local function delRange(esp, esp_type, range)
	range = range or DEFAULT_RANGE
	local range_list = esp[esp_type]

	if range_list then
		local i = 1
		while i <= #range_list and range_list[i] ~= range do i = i + 1 end
		
		if range_list[i] == range then
			for j = i, #range_list do
				range_list[j] = range_list[j + 1]
			end

			if #range_list <= 0 then esp[esp_type] = nil end
		else
			print("ESP error - delRange(): Non-existent range "..range.." for active ESP type "..esp_type)
			return false
		end
	else
		print("ESP error - delRange(): "..esp_type.." is not an active ESP type")
		return false
	end
	
	return true
end

--- Get an effective, modified range.
-- @param esp_range The Actor's ESP range modifier.
-- @param range The range to modify.
-- @return The fully modified range.
local function modifyRange(esp_range, range)
	if not range then return 0 end

	range = math.floor(range + esp_range or 0)
	return range > 0 and range or 0
end


---------------------------------------------------------------------
-- Class methods
---------------------------------------------------------------------

--- Initialize a ESP variables.
-- @param t An optional reference to an object that supports the ESP interface.
function _M:init(t)
	t = t or {}
	self.esp = t.esp or {}
	self.esp_range = t.esp_range or 0
	self.esp_store = t.esp_store or {}
end

--- Add new ESP types and ranges.
-- @param esp_types A table of ESP types and/or ESP type = range pairs OR a string containing a single
--                  ESP type. All types without explicit ranges will be given the default range.
-- @return The number of ESP effects successfully added.
function _M:addEsp(esp_types)
	local esp_types = normalizeEspTypesAndRanges(esp_types)

	return changeTypesAndRanges(self.esp, esp_types, addRange)	
end

--- Remove ESP types and ranges.
-- @param esp_types A table of ESP types and/or ESP type = range pairs OR a string containing a single
--                  ESP type. All types without explicit ranges will assume the default range was used
--                  when it was added.
-- @return The number of ESP effects successfully removed.
function _M:removeEsp(esp_types)
	local esp_types = normalizeEspTypesAndRanges(esp_types)

	return changeTypesAndRanges(self.esp, esp_types, delRange)
end

--- Add ESP types and ranges from temporary sources with automatic storage of effects.
-- @param source An identifier for the source adding the ESP effect.
-- @param esp_types A table of ESP types and/or ESP type = range pairs OR a string containing a single
--                  ESP type. All types without explicit ranges will assume the default range was used
--                  when it was added.
-- @return The number of ESP effects successfully added.
function _M:addTemporaryEsp(source, esp_types)
	self.esp_store[source] = esp_types
	return self:addEsp(esp_types)
end

--- Remove ESP types and ranges from temporary sources using automatic storage of effects.
-- @param source An identifier for the source removing the ESP effect.
-- @return The number of ESP effects successfully removed.
function _M:removeTemporaryEsp(source)
	local ret = 0

	if self.esp_store[source] then
		ret = self:removeEsp(self.esp_store[source])
		self.esp_store[source] = nil
	end

	return ret
end

--- Normalizes an ESP type or type/sub-type pair.
-- @param esp_type The ESP type to normalize.
-- @return The normalized type.
function _M.normalizeEspType(esp_type)
	-- Pop an error() instead of print()ing and continuing since almost all ESP functions start here.
	if type(esp_type) ~= "string" then error("ESP error - normalizeEspType(): Only works with strings") end

	local actor_type, actor_subtype = _M.splitEspType(esp_type)
	local function strip(str)
		return (str:gsub("^[ \t\n\r]*(.-)[ \t\n\r]*$", "%1"))  -- Strips whitespace from string ends
	end

	actor_type = strip(actor_type):lower():capitalize()
	if actor_subtype then
		actor_subtype = strip(actor_subtype):lower():capitalize()
		if actor_subtype ~= "" then actor_type = actor_type.."/"..actor_subtype end
	end
	
	return actor_type
end

--- Join and normalize a type and sub-type.
-- @param actor_type A string containing an ESP type.
-- @param actor_subtype An optional string containing an ESP sub-type.
-- @return A normalized string containing the type/sub-type pair.
function _M.makeEspType(actor_type, actor_subtype)
	actor_type = _M.normalizeEspType(actor_type)
	if actor_subtype then
		actor_subtype = _M.normalizeEspType(actor_subtype)
		if actor_subtype ~= "" then actor_type = actor_type..'/'..actor_subtype end
	end

	return actor_type
end

--- Split a type/sub-type into its parts.
-- @param esp_type A string containing the (possibly non-normal) ESP type or type/sub-type to split,
--                 if applicable.
-- @return A string containing the ESP type and a string with the sub-type, if any.
--         If there is no sub-type, nil is returned instead (i.e. "type", "sub-type" or "type", nil).
function _M.splitEspType(esp_type)
	local slash = (esp_type:find("/"))

	if slash then
		return esp_type:sub(1, slash - 1), esp_type:sub(slash + 1, #esp_type)
	end

	return esp_type
end

--- Get all ESP types and their longest active (not necessarily effective) ranges.
-- @return A table of all ESP type = longest range pairs.
function _M:getActiveEspTypes()
	local esps = {}

	for esp_type, _ in pairs(self.esp) do
		esps[esp_type] = self:getEspRange(esp_type)
	end

	return esps
end

--- Get a table of active ESP matches for a given ESP type.
-- @return A table of all ESP type = range pairs that match a given ESP type.  The table is
--         empty if no types match.
function _M:getMatchingEspTypes(esp_type)
	esp_type = _M.normalizeEspType(esp_type)
	local matches = {}

	for active_type, range_list in pairs(self.esp) do
		if esp_type == active_type or (_M.splitEspType(esp_type)) == active_type or
		   active_type == "All" then
			matches[active_type] = range_list[1]
		end
	end

	return matches
end

--- Get the range of an ESP type.
-- @param esp_type A string containing the ESP type.
-- @return The unmodified range of the requested ESP type.  Ranges are not necessarily integer.
--         Returns nil if the type is not active or has negative range.
function _M:getEspRange(esp_type)
	esp_type = _M.normalizeEspType(esp_type)
	local best_range, matched = 0, false

	for active_type, range in pairs(self:getMatchingEspTypes(esp_type)) do
		if range >= best_range then best_range, matched = range, true end
	end

	return matched and best_range or nil
end

--- Get the effective (modified) range of an ESP type.
-- @param esp_type A string containing the ESP type.
-- @return The primary, modified range of the requested ESP type.  Always integer.  Returns zero (0)
--         if the type is not active.
function _M:getEffectiveEspRange(esp_type)
	return modifyRange(self.esp_range, self:getEspRange(esp_type))
end

--- Get the longest range of any active ESP effect.
-- @return The effective range of the longest reaching ESP effect type.  Will return zero (0)
--         if there are no active ESP effects.  Always integer.
function _M:getMaxEspRange()
	local longest, matched = 0, false
	for _, range in pairs(self:getActiveEspTypes()) do
		if range > longest then longest, matched = range, true end
	end

	return modifyRange(self.esp_range, matched and longest or nil)
end

--- Check if there are any active ESP types.
-- @return Boolean that indicates there is (true) or is not (false) any active ESP.
function _M:hasActiveEsp()
	return (next(self.esp) ~= nil and self:getMaxEspRange() > 0)
end

--- Check if an ESP type can be sensed.
-- @param esp_type A string containing the ESP type.
-- @return A boolean indicating whether it can (true) or cannot (false) be sensed.
function _M:isSensable(esp_type)
	return (self:getEffectiveEspRange(esp_type) > 0)
end

--- Check if one Actor can sense another telepathically.
-- @param actor The Actor being sensed.
-- @return A boolean indicating successful detection.
function _M:canSense(actor)
	local sensed = false

	if self ~= actor and self:hasActiveEsp() and self.x and self.y and actor.x and actor.y then
		local target_esp_type = _M.makeEspType(actor.type, actor.subtype)
		local esp_distance = self:getEffectiveEspRange(target_esp_type)

		-- Check for Allies ESP type.  Only works for Player.
		if self == game.player and self:isSensable("Allies") and (actor:reactionToward(self) > 0 or game.party:hasMember(actor)) then
			esp_distance = math.max(esp_distance, self:getEffectiveEspRange("Allies"))
		end

		if esp_distance >= core.fov.distance(self.x, self.y, actor.x, actor.y, true) then
			sensed = true
		end
	end

	return sensed
end

--- Create a comparison table sort function.
-- @param order_table An array containing ESP types to list first, in order.
-- @return A function that sorts a comparison table, order_table elements first (in order_table
--         element order), and the rest alphabetically.
function _M.makeEspComparisonSorter(order_table)
	order_table = order_table or {"All"}
	sort_str = "return function (l, r) return "

	for _, esp_type in ipairs(order_table) do
		esp_type = _M.normalizeEspType(esp_type)
		sort_str = sort_str .. "(l[1] == \""..esp_type.."\" or r[1] ~= \""..esp_type.."\" and "
	end

	sort_str = sort_str .. "l[1] < r[1]"

	for i = 1, #order_table do
		sort_str = sort_str .. ")"
	end

	sort_str = sort_str .. " end"

	return loadstring(sort_str)()
end

--- Generate a sorted comparison table from two ESP objects.
-- @param left An ESP table.
-- @param right Optionally, another ESP table.
-- @param sorter A function to sort ESP types as desired.
-- @return A comparison table of the two.
-- Comparison tables are defined as follows:
--	Each element of the table is an array that has an ESP type as [1] and another table as [2].
--  The tables in [2] are also arrays of two elements with left's range in [1] and right's in [2].
--  If one of the object's lacks that ESP type then its array index will be nil instead of a range.
--  The "All" ESP type will be the first element (if either left or right has that type) and the
--  rest will be in alphabetical order.  That is:
--  {{leftonly, {1, nil}}, {rightonly, {nil, 1}}, {both = {1, 1}}}
_M.defaultEspComparisonSorter = _M.makeEspComparisonSorter{"All", "Allies"}
function _M.makeEspComparisonTable(left, right, sorter)
	local compare_table = {}
	right = right or {}
	sorter = sorter or _M.defaultEspComparisonSorter

	-- Insert left ESP types into sortable list
	for esp_type, ranges in pairs(left) do
		table.insert(compare_table, {esp_type, {ranges[1]}})
	end

	-- Insert right ESP types into sortable list
	for esp_type, ranges in pairs(right) do
		local found = false
		for _, esp in ipairs(compare_table) do
			if esp[1] == esp_type then
				esp[2][2] = ranges[1]
				found = true
				break
			end
		end
		if not found then
			table.insert(compare_table, {esp_type, {nil, ranges[1]}})
		end
	end

	-- Sort the list of ESP types using the supplied sort routine
	table.sort(compare_table, sorter)
	
	return compare_table
end

return require("engine.class").make(_M)